@@ -1,5 +1,5 @@
 #
-# $Id: Layer.pm 360 2015-01-20 18:36:06Z gomor $
+# $Id: Layer.pm 363 2015-01-28 06:53:20Z gomor $
 #
 package Net::Frame::Layer;
 use strict;
@@ -95,51 +95,115 @@ sub dump { CORE::unpack('H*', shift->raw) }
 # Useful subroutines
 #
 
-# load AF_INET and default imports from Socket.  Safe back to at least 5.8.8
+# Load AF_INET and default imports from Socket. Safe back to at least 5.8.8.
 use Socket qw(:DEFAULT AF_INET);
+
+sub _setInet6Sub {
+   no strict 'refs';
+
+   # Check Socket against some IPv6 functions and constants.
+   eval {
+      require Socket;
+      Socket->import(qw(AF_INET6 inet_pton inet_ntop));
+   };
+   if (! $@) {   # Socket has support for required functions and constants.
+      *{__PACKAGE__.'::_inet_pton'} = \&Socket::inet_pton;
+      *{__PACKAGE__.'::_inet_ntop'} = \&Socket::inet_ntop;
+
+      return 1;
+   }
+
+   # Fallback to Socket6
+   eval {
+      require Socket6;
+      Socket6->import(qw(AF_INET6 inet_pton inet_ntop));
+   };
+   if (! $@) {   # Socket6 has support for required functions and constants.
+      *{__PACKAGE__.'::_inet_pton'} = \&Socket6::inet_pton;
+      *{__PACKAGE__.'::_inet_ntop'} = \&Socket6::inet_ntop;
+
+      return 1;
+   }
+
+   die("[-] Net::Frame: inet_pton/inet_ntop: not supported by Socket nor Socket6: ".
+       "try upgrading your Perl version or Socket/Socket6 modules.\n");
+}
+
+sub _setGetAddressSub {
+   no strict 'refs';
+
+   # Check Socket against some IPv6 functions and constants.
+   eval {
+      require Socket;
+      Socket->import(qw(getaddrinfo getnameinfo AF_INET6 NI_NUMERICHOST NI_NUMERICSERV));
+   };
+   if (! $@) {   # Socket has support for required functions and constants.
+      *{__PACKAGE__.'::_getAddress'} = sub {
+         my ($name) = @_;
+
+         #print STDERR "*** Socket supports IPv6 OK\n";
+
+         my %hints = (
+            family => Socket::AF_INET6(),
+            protocol => SOCK_STREAM,
+         );
+         my ($err, @res) = Socket::getaddrinfo($name, '', \%hints);
+         if ($err) {
+            carp("@{[(caller(0))[3]]}: unable to resolv `$name' hostname: getaddrinfo: $err\n");
+            return;
+         }
+         if (@res > 0) {
+            my $h = $res[0];
+            my ($err, $ipv6) = Socket::getnameinfo(
+               $h->{addr}, Socket::NI_NUMERICHOST() | Socket::NI_NUMERICSERV()
+            );
+            if ($err) {
+               carp("@{[(caller(0))[3]]}: unable to resolv `$name' hostname: getnameinfo: $err\n");
+               return;
+            }
+
+            return $ipv6;
+         }
+         else {
+            carp("@{[(caller(0))[3]]}: unable to resolv `$name' hostname: getaddrinfo: $!\n");
+            return;
+         }
+      };
+
+      return 1;
+   }
+
+   # Fallback to Socket6
+   eval {
+      require Socket6;
+      Socket6->import(qw(getaddrinfo getnameinfo AF_INET6 NI_NUMERICHOST NI_NUMERICSERV));
+   };
+   if (! $@) {   # Socket6 has support for required functions and constants.
+      *{__PACKAGE__.'::_getAddress'} = sub {
+         my ($name) = @_;
+
+         #print STDERR "*** Fallback to Socket6 support\n";
+
+         my @res = Socket6::getaddrinfo($name, '', Socket6::AF_INET6(), SOCK_STREAM);
+         if (@res >= 5) {
+            my ($ipv6) = Socket6::getnameinfo(
+               $res[3], Socket6::NI_NUMERICHOST() | Socket6::NI_NUMERICSERV()
+            );
+
+            return $ipv6;
+         }
+      };
+
+      return 1;
+   }
+
+   die("[-] Net::Frame: getaddrinfo/getnameinfo: not supported by Socket nor Socket6: ".
+       "try upgrading your Perl version or Socket/Socket6 modules.\n");
+}
+
 BEGIN {
-    # imports that may or may not be in Socket.
-    my @imports = (
-        qw(AF_INET6 NI_NUMERICHOST NI_NUMERICSERV getaddrinfo getnameinfo));
-    my @socket6_imports;
-
-    # This might be overkill, but I'm not certain that all these imports
-    # were added to Socket at the same time.
-    for my $export (@imports) {
-        eval { Socket->import($export); };
-        if ($@) {
-            push @socket6_imports, $export;
-        }
-    }
-
-    eval {
-        # Test to see if the sub works.
-        # Socket::inet_ntop() may exist, but die with:
-        # Socket::inet_ntop not implemented on this architecture
-        Socket::inet_ntop( AF_INET, "\0\0\0\0" );    # test
-        Socket->import('inet_ntop');    # import if the test doesn't die
-    };
-    if ($@) {
-        push @socket6_imports, 'inet_ntop';
-    }
-    eval {
-        # Test to see if the sub works.
-        # Socket::inet_pton() may exist, but die with:
-        # Socket::inet_pton not implemented on this architecture
-        Socket::inet_pton( AF_INET, '0.0.0.0' );    # test
-        Socket->import('inet_pton');    # import if the test doesn't die
-    };
-    if ($@) {
-        push @socket6_imports, 'inet_pton';
-    }
-
-    if (@socket6_imports) {
-
-        # something we want wasn't found in Socket time to try Socket6
-        eval { require Socket6 };
-        die $@ if $@;
-        Socket6->import(@socket6_imports);
-    }
+   _setInet6Sub();
+   _setGetAddressSub();
 }
 
 require Net::IPv6Addr;
@@ -147,49 +211,73 @@ require Net::IPv6Addr;
 sub getHostIpv4Addr {
    my ($name) = @_;
 
-   return undef unless $name;
-   return $name if $name =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
+   # No address given
+   if (! defined($name)) {
+      return;
+   }
+
+   # Already an IPv4 address
+   if ($name =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
+      return $name;
+   }
 
    my @addrs = (gethostbyname($name))[4];
    @addrs ? return join('.', CORE::unpack('C4', $addrs[0]))
           : carp("@{[(caller(0))[3]]}: unable to resolv `$name' hostname\n");
-   return undef;
+
+   # Error
+   return;
 }
 
 sub getHostIpv4Addrs {
    my ($name) = @_;
 
-   return undef unless $name;
-   return $name if $name =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
+   # No address given
+   if (! defined($name)) {
+      return;
+   }
+
+   # Already an IPv4 address
+   if ($name =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
+      return $name;
+   }
 
    my @addrs = (gethostbyname($name))[4];
    @addrs ? return @addrs
           : carp("@{[(caller(0))[3]]}: unable to resolv `$name' hostname\n");
+
+   # Error
    return ();
 }
 
 sub getHostIpv6Addr {
    my ($name) = @_;
 
-   return undef unless $name;
-   return $name if Net::IPv6Addr::is_ipv6($name);
+   # No address given
+   if (! defined($name)) {
+      return;
+   }
 
-   my @res = getaddrinfo($name, 'ssh', AF_INET6, SOCK_STREAM);
-   if (@res >= 5) {
-      my ($ipv6) = getnameinfo($res[3], NI_NUMERICHOST | NI_NUMERICSERV);
-      $ipv6 =~ s/%.*$//;
-      return $ipv6;
+   # Already an IPv6 address
+   if (Net::IPv6Addr::is_ipv6($name)) {
+      return $name;
    }
-   else {
+
+   my $ipv6 = _getAddress($name);
+   if (! defined($ipv6)) {
       carp("@{[(caller(0))[3]]}: unable to resolv `$name' hostname\n");
+      return;
    }
-   undef;
+
+   $ipv6 =~ s/%.*$//;
+
+   return $ipv6;
 }
 
-sub inetAton  { inet_aton(shift()) }
-sub inetNtoa  { inet_ntoa(shift()) }
-sub inet6Aton { inet_pton(AF_INET6, shift()) }
-sub inet6Ntoa { inet_ntop(AF_INET6, shift()) }
+sub inetAton  { Socket::inet_aton(shift()) }
+sub inetNtoa  { Socket::inet_ntoa(shift()) }
+sub inet6Aton { _inet_pton(AF_INET6, shift()) }
+sub inet6Ntoa { _inet_ntop(AF_INET6, shift()) }
 
 sub getRandomHighPort {
    my $highPort = int rand 0xffff;
